ReforgeSummary = LibStub("AceAddon-3.0"):NewAddon("ReforgeSummary", "AceEvent-3.0", "AceHook-3.0")
local core = ReforgeSummary
local ReforgingInfo = LibStub("LibReforgingInfo-1.0")
local Gratuity = LibStub("LibGratuity-3.0")

local arrow = "|TInterface\\AddOns\\ReforgeSummary\\arrow.tga:0:1.15:0:1:64:64:17:47:19:45|t"
local displayOrder = {
    INVSLOT_HEAD,
    INVSLOT_NECK,
    INVSLOT_SHOULDER,
    INVSLOT_BACK,
    INVSLOT_CHEST,
    INVSLOT_WRIST,

    INVSLOT_HAND,
    INVSLOT_WAIST,
    INVSLOT_LEGS,
    INVSLOT_FEET,
    INVSLOT_FINGER1,
    INVSLOT_FINGER2,
    INVSLOT_TRINKET1,
    INVSLOT_TRINKET2,

    INVSLOT_MAINHAND,
    INVSLOT_OFFHAND,
    INVSLOT_RANGED,
}
-- Mapped from reforging stat ids. Not using LibReforgingInfo because "rating"
-- is appended on hit and crit
local extraShort = {
	ITEM_MOD_SPIRIT_SHORT,
	STAT_DODGE,
	STAT_PARRY,
    HIT,
    CRIT_ABBR,
	STAT_HASTE,
	STAT_EXPERTISE,
	STAT_MASTERY,
}
-- Mapped from reforging stat ids to tooltip label to scan for
local scanLine = {
    ITEM_MOD_SPIRIT,
    ITEM_MOD_DODGE_RATING,
    ITEM_MOD_PARRY_RATING,
    ITEM_MOD_HIT_RATING,
    ITEM_MOD_CRIT_RATING,
    ITEM_MOD_HASTE_RATING,
    ITEM_MOD_EXPERTISE_RATING,
    ITEM_MOD_MASTERY_RATING,
}
-- Replace any occurance of %d with a number capture
-- also replace any period with an escaped version
-- also replace any %c ("+" in +50 Spirit) with a single .
for i,str in ipairs(scanLine) do
    scanLine[i] = str:gsub("%.", "%%."):gsub("%%d", "(%%d+)"):gsub("%%c", ".")
end
-- Map for stat ids returned by GetReforgeOptionInfo()
local statMap = {
    [6]  = 1, -- Spirit
    [13] = 2, -- Dodge
    [14] = 3, -- Parry
    [31] = 4, -- Hit
    [32] = 5, -- Crit
    [36] = 6, -- Haste
    [37] = 7, -- Expertise
    [49] = 8, -- Mastery
}
-- GameTooltip methods that we want to overwrite with additional tooltip information
-- The mapped value should return the reforge details
local tooltipMethods = {
    SetBagItem              = function(container, slot) return core:ParseHyperlink(GetContainerItemLink(container, slot)) end,
    SetHyperlink            = function(hyperlink)       return core:ParseHyperlink(hyperlink) end,
    SetHyperlinkCompareItem = function(hyperlink)       return core:ParseHyperlink(hyperlink) end,
    SetInventoryItem        = function(unit, slot)      return core:ParseHyperlink(GetInventoryItemLink(unit, slot)) end,
    SetReforgeItem          = function()
        local r = GetReforgeItemInfo()
        if r and r > 0 then
            local _, src, amt, _, dst = GetReforgeOptionInfo(r)
            return amt, statMap[src], statMap[dst]
        end
    end,
    -- SetTransmogrifyItem     = function() end, -- TODO ?
}
local tooltipFrames = {
    GameTooltip,
    ItemRefTooltip,
    ShoppingTooltip1,
    ShoppingTooltip2,
    ShoppingTooltip3,
}

    
local anchor = CreateFrame("Frame") -- Anchor for first item in display list
local buttons = {} -- maps inventory slot to frame
local pre43 = select(4, GetBuildInfo()) < 40300 -- pre 4.3 clients will have to hack up the frame a bit

function core:OnEnable()
    -- Hook tooltips
    for i, frame in ipairs(tooltipFrames) do
        for method, filter in pairs(tooltipMethods) do
            self:SecureHook(frame, method, function(...) self:AlterTooltip(filter, ...) end)
        end
    end

    -- Hook reforging ui when loaded
    if IsAddOnLoaded("Blizzard_ReforgingUI") then
        self:Setup()
    else
        self:RegisterEvent("ADDON_LOADED", function(_, name)
            if name == "Blizzard_ReforgingUI" then
                self:Setup()
                self:UnregisterEvent("ADDON_LOADED")
            end
        end)
    end
end

function core:Setup()
    -- Prevent item pickup when clicking after reforge
    self:SecureHook("ReforgingFrame_AddItemClick", function() ClearCursor() end)

    -- Right click on frame while in reforger will back out
    self:HookScript(ReforgingFrame, "OnMouseUp", function(self, button) if button == "RightButton" then ReforgingFrame_AddItemClick() end end)

    -- Keep the glow animation, but trigger the change immediately instead of OnFinish
    local anim
    if pre43 then
        anim = ReforgingFrame.topInset.invisButton.glow.reforgeAnim
    else
        anim = ReforgingFrame.glow.reforgeAnim
    end
    local finished = anim:GetScript("OnFinished", nil)
    self:SecureHook(anim, "Play", finished)
    anim:SetScript("OnFinished", nil)

    -- Check when to show
    self:SecureHook("ReforgingFrame_Update", "CheckUpdate")

    -- Update on inventory change
    self:RegisterEvent("UNIT_INVENTORY_CHANGED", "CheckUpdate")

    -- Run Update
    self:CheckUpdate()

    if pre43 then
        ReforgingFrame:SetWidth(400)
    end
end

function core:CheckUpdate()
    if ReforgingFrame.missingDescription:IsVisible() then
        -- Hide 4.2 art
        if pre43 then
            ReforgingFrameBottomBg:Hide()
            ReforgingFrameBottomInset:Hide()
            ReforgingFrameInset:SetPoint("BOTTOMRIGHT", ReforgingFrameReforgeButton, "TOPRIGHT", 0, 4)
        end

        -- Hide gray text
        ReforgingFrame.missingDescription:SetText("")

        -- Check for init
        if not next(buttons) then
            self:CreateButtons()
        end

        -- Show all items
        self:Buttons("Update")
    else
        -- Show 4.2 art
        if pre43 then
            ReforgingFrameBottomBg:Show()
            ReforgingFrameBottomInset:Show()
            ReforgingFrameInset:SetPoint("BOTTOMRIGHT", ReforgingFrameBottomInset, "TOPRIGHT", 0, 4)
        end

        -- Hide all items
        self:Buttons("Hide")
    end
end

function core:Buttons(func, ...)
    for _, frame in pairs(buttons) do
        frame[func](frame, ...)
    end
end

function core:CreateButtons()
    -- Position anchor
    anchor:ClearAllPoints()
    anchor:SetPoint("TOPLEFT", ReforgingFrameHorzBar, "TOPLEFT", 0, 15)
    anchor:SetSize(1, 1)

    local last = anchor
    local width = 380
    local font = GameFontNormal

    for _,slot in ipairs(displayOrder) do
        local frame = CreateFrame("Button")
        frame:SetParent(ReforgingFrame)
        frame:SetSize(width, 14)
        frame:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -1)
        frame:RegisterForClicks("LeftButtonUp")
        
        -- Item icon
        local icon = CreateFrame("Frame")
        icon:SetParent(frame)
        icon:SetSize(14, 14)
        icon:SetPoint("LEFT")
        local iconTexture = icon:CreateTexture()
        iconTexture:SetAllPoints()

        -- Item name
        local name = frame:CreateFontString()
        name:SetFontObject(font)
        name:SetSize(170, 14)
        name:SetJustifyH("LEFT")
        name:SetPoint("LEFT", icon, "RIGHT", 3, 0)

        -- Stat amount
        local refAmt = frame:CreateFontString()
        refAmt:SetFontObject(font)
        refAmt:SetSize(45, 14)
        refAmt:SetJustifyH("RIGHT")
        refAmt:SetPoint("LEFT", icon, "RIGHT", 175, 0)

        -- Stat label from
        local fromStat = frame:CreateFontString()
        fromStat:SetFontObject(font)
        fromStat:SetSize(65, 14)
        fromStat:SetJustifyH("LEFT")
        fromStat:SetPoint("LEFT", icon, "RIGHT", 225, 0)

        -- Stat label to
        local toStat = frame:CreateFontString()
        toStat:SetFontObject(font)
        toStat:SetSize(65, 14)
        toStat:SetJustifyH("LEFT")
        toStat:SetPoint("LEFT", icon, "RIGHT", 300, 0)

        -- Handlers
        frame:SetScript("OnClick", function(self)
            PickupInventoryItem(slot)
            SetReforgeFromCursorItem()
        end)
        frame:SetScript("OnEnter", function(self)
            GameTooltip:SetOwner(self, "ANCHOR_BOTTOMLEFT", width, -16)
            GameTooltip:SetInventoryItem("player", slot)
            GameTooltip:Show()

            name:SetTextColor(1, 1, 1, 1)
            refAmt:SetTextColor(1, 1, 1, 1)
            fromStat:SetTextColor(1, 1, 1, 1)
            toStat:SetTextColor(1, 1, 1, 1)
        end)
        frame:SetScript("OnLeave", function(self)
            GameTooltip:Hide()

            name:SetTextColor(font:GetTextColor())
            refAmt:SetTextColor(font:GetTextColor())
            fromStat:SetTextColor(font:GetTextColor())
            toStat:SetTextColor(font:GetTextColor())
        end)

        -- Set info for this button
        frame.Update = function(self)
            local link = GetInventoryItemLink("player", slot)
            if not link then
                self:Hide()
                return
            end

            self:Show()
            -- Update name and icon
            name:SetText(link:match("%[(.+)%]"))
            iconTexture:SetTexture(GetInventoryItemTexture("player", slot))

            -- Get stats
            local amt, from, to = core:ParseHyperlink(link)
            if not amt then
                -- Not reforged, clear it
                refAmt:SetText("")
                fromStat:SetText("")
                toStat:SetText("")
            else
                -- Set stat labels
                refAmt:SetText(amt)
                fromStat:SetText(extraShort[from])
                toStat:SetText(extraShort[to])
            end
        end

        last = frame
        buttons[slot] = frame
    end
end

function core:ParseHyperlink(link)
    if not link then return end
    -- Get stats
    local stats = ReforgingInfo:GetReforgeID(link)
    -- Not reforged
    if not stats then
        return nil
    end

    -- Get stat ids
    local from, to = ReforgingInfo:GetReforgedStatIDs(stats)

    -- Scan tooltip for amount
    Gratuity:SetHyperlink(link)
    local pattern = scanLine[to]
    local _, _, amt = Gratuity:Find(pattern)
    if not amt then
        -- error("Could not find " .. scanLine[to] .. " in " .. link)
        amt = 0
    end

    return amt, from, to
end

function core:AlterTooltip(filter, tooltip, ...)
    -- Use filters in tooltip map and parameters of calling function to get
    -- reforge info
    local amt, from, to = filter(...)
    if not amt then return end

    -- Scan tooltip for the "Reforged" line
    for i=1, tooltip:NumLines() do
        local text = _G[tooltip:GetName() .. "TextLeft" .. i]
        if text:GetText() == REFORGED then
            text:SetFormattedText("%s (%d %s %s %s)", REFORGED, amt, extraShort[from], arrow, extraShort[to])
            break
        end
    end
end
